"""Inspired by https://github.com/pygae/clifford"""

import functools
import itertools
import operator

import torch


# copied from the itertools docs
def _powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


class ShortLexBasisBladeOrder:
    def __init__(self, n_vectors):
        self.index_to_bitmap = torch.empty(2**n_vectors, dtype=int)
        self.grades = torch.empty(2**n_vectors, dtype=int)
        self.bitmap_to_index = torch.empty(2**n_vectors, dtype=int)

        for i, t in enumerate(_powerset([1 << i for i in range(n_vectors)])):
            bitmap = functools.reduce(operator.or_, t, 0)
            self.index_to_bitmap[i] = bitmap
            self.grades[i] = len(t)
            self.bitmap_to_index[bitmap] = i
            del t  # enables an optimization inside itertools.combinations


def set_bit_indices(x: int):
    """Iterate over the indices of bits set to 1 in `x`, in ascending order"""
    n = 0
    while x > 0:
        if x & 1:
            yield n
        x = x >> 1
        n = n + 1


def count_set_bits(bitmap: int) -> int:
    """Counts the number of bits set to 1 in bitmap"""
    count = 0
    for i in set_bit_indices(bitmap):
        count += 1
    return count


def canonical_reordering_sign_euclidean(bitmap_a, bitmap_b):
    """
    Computes the sign for the product of bitmap_a and bitmap_b
    assuming a euclidean metric
    """
    a = bitmap_a >> 1
    sum_value = 0
    while a != 0:
        sum_value = sum_value + count_set_bits(a & bitmap_b)
        a = a >> 1
    if (sum_value & 1) == 0:
        return 1
    else:
        return -1


def canonical_reordering_sign(bitmap_a, bitmap_b, metric):
    """
    Computes the sign for the product of bitmap_a and bitmap_b
    given the supplied metric
    """
    bitmap = bitmap_a & bitmap_b
    output_sign = canonical_reordering_sign_euclidean(bitmap_a, bitmap_b)
    i = 0
    while bitmap != 0:
        if (bitmap & 1) != 0:
            output_sign *= metric[i]
        i = i + 1
        bitmap = bitmap >> 1
    return output_sign


def gmt_element(bitmap_a, bitmap_b, sig_array):
    """
    Element of the geometric multiplication table given blades a, b.
    The implementation used here is described in :cite:`ga4cs` chapter 19.
    """
    output_sign = canonical_reordering_sign(bitmap_a, bitmap_b, sig_array)
    output_bitmap = bitmap_a ^ bitmap_b
    return output_bitmap, output_sign


# def construct_gmt(index_to_bitmap, bitmap_to_index, signature):
#     n = len(index_to_bitmap)
#     array_length = int(n * n)
#     coords = torch.zeros((3, array_length), dtype=torch.uint8)
#     k_list = coords[0, :]
#     l_list = coords[1, :]
#     m_list = coords[2, :]

#     # use as small a type as possible to minimize type promotion
#     mult_table_vals = torch.zeros(array_length)
#     inner_table_vals = torch.zeros(array_length)
#     outer_table_vals = torch.zeros(array_length)

#     for i in range(n):
#         bitmap_i = index_to_bitmap[i]

#         for j in range(n):
#             bitmap_j = index_to_bitmap[j]
#             bitmap_v, mul = gmt_element(bitmap_i, bitmap_j, signature)
#             v = bitmap_to_index[bitmap_v]

#             list_ind = i * n + j
#             k_list[list_ind] = i
#             l_list[list_ind] = v
#             m_list[list_ind] = j

#             mult_table_vals[list_ind] = mul
#             if v + 1 == abs(i - j):
#                 inner_table_vals[list_ind] = mul
#             if v + 1 == i + j + 2:
#                 outer_table_vals[list_ind] = mul
                

#     return torch.sparse_coo_tensor(indices=coords, values=mult_table_vals, size=(n, n, n)), \
#             torch.sparse_coo_tensor(indices=coords, values=inner_table_vals, size=(n, n, n)), \
#             torch.sparse_coo_tensor(indices=coords, values=outer_table_vals, size=(n, n, n))

import numpy as np
def construct_gmt(index_to_bitmap, bitmap_to_index, signature):
    n = len(index_to_bitmap)
    array_length = int(n * n)

    t_geom = np.zeros((n, n, n), dtype=np.int32)
    t_inner = np.zeros((n, n, n), dtype=np.int32)
    t_outer = np.zeros((n, n, n), dtype=np.int32)
    
    
    for i in range(n):
        bitmap_i = index_to_bitmap[i]

        for j in range(n):
            bitmap_j = index_to_bitmap[j]
            bitmap_v, mul = gmt_element(bitmap_i, bitmap_j, signature)
            v = bitmap_to_index[bitmap_v]

            t_geom[i,j,v] = mul
            if v+1 == abs(i - j):
                t_inner[i,j,v] = mul
            if v+1 == i + j +2:
                t_outer[i,j,v] = mul
                
    return  [torch.tensor(_) for _ in [t_geom,t_inner,t_outer]]
        


"""based on https://github.com/DavidRuhe/clifford-group-equivariant-neural-networks"""

import functools
import math

import torch
from torch import nn

from torch_ga.clifford.blades import ShortLexBasisBladeOrder, construct_gmt, gmt_element
from torch_ga import GeometricAlgebra, MultiVector

class CliffordAlgebra(nn.Module):
    def __init__(self, metric):
        super().__init__()

        self.register_buffer("metric", torch.as_tensor(metric.detach() if isinstance(metric, torch.Tensor) else metric))
        self.num_bases = len(metric)
        self.bbo = ShortLexBasisBladeOrder(self.num_bases)
        self.dim = len(self.metric)
        self.n_blades = len(self.bbo.grades)
        cayley, cayley_inner, cayley_outer  = [_ for _ in construct_gmt(self.bbo.index_to_bitmap, self.bbo.bitmap_to_index, self.metric)]
        cayley, cayley_inner, cayley_outer  = [ _.to_dense().to(torch.get_default_dtype()) for _ in [cayley, cayley_inner, cayley_outer]] 
        self.grades = self.bbo.grades.unique()
        self.register_buffer(
            "subspaces",
            torch.tensor(tuple(math.comb(self.dim, g) for g in self.grades)),
        )
        self.n_subspaces = len(self.grades)
        self.grade_to_slice = self._grade_to_slice(self.subspaces)
        self.grade_to_index = [
            torch.tensor(range(*s.indices(s.stop))) for s in self.grade_to_slice
        ]

        self.register_buffer(
            "bbo_grades", self.bbo.grades.to(torch.get_default_dtype())
        )
        self.register_buffer("even_grades", self.bbo_grades % 2 == 0)
        self.register_buffer("odd_grades", ~self.even_grades)
        self.register_buffer("cayley", cayley)
        self.register_buffer("cayley_inner", cayley_inner)
        self.register_buffer("cayley_outer", cayley_outer)
        
    def geometric_product(self, a, b, blades=None):
        return self.product(a, b, blades, _type="geometric")
    def inner_product(self, a, b, blades=None):
        return self.product(a, b, blades, _type="inner")
    def outer_product(self, a, b, blades=None):
        return self.product(a, b, blades, _type="outer")
    def product(self, a, b, blades=None, _type="geometric"):
        if _type in ["geometric"]:
            cayley = self.cayley
        elif _type in ["inner"]:
            cayley = self.cayley
        elif _type in ["outer"]:
            cayley = self.cayley_outer
        else:
            raise Exception(f"Unknownd product {_type}")

        if blades is not None:
            blades_l, blades_o, blades_r = blades
            assert isinstance(blades_l, torch.Tensor)
            assert isinstance(blades_o, torch.Tensor)
            assert isinstance(blades_r, torch.Tensor)
            cayley = cayley[blades_l[:, None, None], blades_o[:, None], blades_r]

        return torch.einsum("...i,ijk,...k->...j", a, cayley, b)

    def __call__(self, a: torch.Tensor) -> MultiVector:
        """Creates a `MultiVector` from a geometric algebra tensor.
        Mainly used as a wrapper for the algebra's functions for convenience.

        Args:
            a: Geometric algebra tensor to return `MultiVector` for

        Returns:
            `MultiVector` for `a`
        """
        if False: a = a.to(dtype=torch.float32)
        # return MultiVector(a, GeometricAlgebra(self.metric.detach().numpy()))
        return MultiVector(a, GeometricAlgebra(self.metric))
    
    def to_ga(self) -> GeometricAlgebra:
        """Creates a `GeometricAlgebra` from the Clifford Algebra

        Returns:
            `GeometricAlgebra` 
        """
        return GeometricAlgebra(self.metric)
    
    def _grade_to_slice(self, subspaces):
        grade_to_slice = list()
        subspaces = torch.as_tensor(subspaces)
        for grade in self.grades:
            index_start = subspaces[:grade].sum()
            index_end = index_start + math.comb(self.dim, grade)
            grade_to_slice.append(slice(index_start, index_end))
        return grade_to_slice

    @functools.cached_property
    def _alpha_signs(self):
        return torch.pow(-1, self.bbo_grades)

    @functools.cached_property
    def _beta_signs(self):
        return torch.pow(-1, self.bbo_grades * (self.bbo_grades - 1) // 2)

    @functools.cached_property
    def _gamma_signs(self):
        return torch.pow(-1, self.bbo_grades * (self.bbo_grades + 1) // 2)

    def alpha(self, mv, blades=None):
        """Clifford main involution"""
        
        signs = self._alpha_signs
        if blades is not None:
            signs = signs[blades]
        return signs * mv.clone()

    def beta(self, mv, blades=None):
        """Clifford main anti-involution (reversion)"""
        signs = self._beta_signs
        if blades is not None:
            signs = signs[blades]
        return signs * mv.clone()

    def gamma(self, mv, blades=None):
        """Clifford conjugation"""
        signs = self._gamma_signs
        if blades is not None:
            signs = signs[blades]
        return signs * mv.clone()

    def zeta(self, mv):
        return mv[..., :1]

    def embed(self, tensor: torch.Tensor, tensor_index: torch.Tensor) -> torch.Tensor:
        mv = torch.zeros(
            *tensor.shape[:-1], 2**self.dim, device=tensor.device, dtype=tensor.dtype
        )
        mv[..., tensor_index] = tensor
        mv = mv.to(dtype=torch.float32)
        return mv

    def embed_grade(self, tensor: torch.Tensor, grade: int) -> torch.Tensor:
        mv = torch.zeros(*tensor.shape[:-1], 2**self.dim, device=tensor.device)
        s = self.grade_to_slice[grade]
        mv[..., s] = tensor
        return mv

    def get(self, mv: torch.Tensor, blade_index: tuple[int]) -> torch.Tensor:
        blade_index = tuple(blade_index)
        return mv[..., blade_index]

    def get_grade(self, mv: torch.Tensor, grade: int) -> torch.Tensor:
        s = self.grade_to_slice[grade]
        return mv[..., s]

    def b(self, x, y, blades=None):
        """Bilinear Form b(x,y): xy+yx = 2b(x,y), I would call the inner product

        Args:
            x (_type_): _description_
            y (_type_): _description_
            blades (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        if blades is not None:
            assert len(blades) == 2
            beta_blades = blades[0]
            blades = (
                blades[0],
                torch.tensor([0]),
                blades[1],
            )
        else:
            blades = torch.tensor(range(self.n_blades))
            blades = (
                blades,
                torch.tensor([0]),
                blades,
            )
            beta_blades = None

        return self.geometric_product(
            self.beta(x, blades=beta_blades),
            y,
            blades=blades,
        )

    def q(self, mv, blades=None):
        """Scalar Square b(x), defines the bilinear form 2 b(x,y) = q(x+y)- q(x) -q(y)

        Args:
            mv (tensor): _description_
            blades (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        if blades is not None:
            blades = (blades, blades)
        return self.b(mv, mv, blades=blades)

    def _smooth_abs_sqrt(self, input, eps=1e-16):
        return (input**2 + eps) ** 0.25

    def norm(self, mv, blades=None):
        return self._smooth_abs_sqrt(self.q(mv, blades=blades))

    def norms(self, mv, grades=None):
        if grades is None:
            grades = self.grades
        return [
            self.norm(self.get_grade(mv, grade), blades=self.grade_to_index[grade])
            for grade in grades
        ]

    def qs(self, mv, grades=None):
        if grades is None:
            grades = self.grades
        return [
            self.q(self.get_grade(mv, grade), blades=self.grade_to_index[grade])
            for grade in grades
        ]

    def reverse(self, mv, blades=None):
        """Perform the reversion operation on multivectors, an operation specific to geometric algebra.

        In Geometric Algebra, the reverse of a multivector is formed by reversing the order of the vectors in each blade.

        Args:
            mv (torch.Tensor): Input multivectors.
            blades (Union[tuple, list, torch.Tensor], optional): Specify which blades are present in the multivector.

        Returns:
            torch.Tensor: The reversed multivector.
        """
        grades = self.bbo.grades.to(mv.device)
        if blades is not None:
            grades = grades[torch.as_tensor(blades, dtype=int)]
        signs = torch.pow(-1, torch.floor(grades * (grades - 1) / 2))
        return signs * mv.clone()
    
    def sandwich(self, u, v, w=None):
        if w is None:
            return self.sandwich2(u,v)
        else:
            return self.sandwich3(u,v,w)
                
    def sandwich2(self, a, b):
        """aba'"""
        return self.geometric_product(self.geometric_product(a, b), self.reverse(a))
    
    def sandwich3(self, u, v, w):
        return self.geometric_product(self.geometric_product(u, v), w)

    def output_blades(self, blades_left, blades_right):
        blades = []
        for blade_left in blades_left:
            for blade_right in blades_right:
                bitmap_left = self.bbo.index_to_bitmap[blade_left]
                bitmap_right = self.bbo.index_to_bitmap[blade_right]
                bitmap_out, _ = gmt_element(bitmap_left, bitmap_right, self.metric)
                index_out = self.bbo.bitmap_to_index[bitmap_out]
                blades.append(index_out)

        return torch.tensor(blades)

    def random(self, n=None):
        if n is None:
            n = 1
        return torch.randn(n, self.n_blades)

    def random_vector(self, n=None):
        if n is None:
            n = 1
        vector_indices = self.bbo_grades == 1
        v = torch.zeros(n, self.n_blades, device=self.cayley.device)
        v[:, vector_indices] = torch.randn(
            n, vector_indices.sum(), device=self.cayley.device
        )
        return v

    def parity(self, mv):
        is_odd = torch.all(mv[..., self.even_grades] == 0)
        is_even = torch.all(mv[..., self.odd_grades] == 0)

        if is_odd ^ is_even:  # exclusive or (xor)
            return is_odd
        else:
            raise ValueError("This is not a homogeneous element.")

    def eta(self, w):
        """Coboundary of alpha"""
        return (-1) ** self.parity(w)

    def alpha_w(self, w, mv):
        return self.even_grades * mv + self.eta(w) * self.odd_grades * mv

    def inverse(self, mv, blades=None):
        mv_ = self.beta(mv, blades=blades)
        return mv_ / self.q(mv)

    def rho(self, w, mv):
        """Applies the versor w action to mv.
        
        Reflects x in the hyperplane normal to w
        
        rho(w)(x) = -wxw^-1 = x-e b(w,x)/b(w,w)w
        
        """
        
        return self.sandwich3(w, self.alpha_w(w, mv), self.inverse(w))

    def reduce_geometric_product(self, inputs):
        return functools.reduce(self.geometric_product, inputs)

    def versor(self, order=None, normalized=True):
        if order is None:
            order = self.dim if self.dim % 2 == 0 else self.dim - 1
        vectors = self.random_vector(order)
        versor = self.reduce_geometric_product(vectors[:, None])
        if normalized:
            versor = versor / self.norm(versor)[..., :1]
        return versor

    def rotor(self):
        return self.versor()

    @functools.cached_property
    def geometric_product_paths(self):
        gp_paths = torch.zeros((self.dim + 1, self.dim + 1, self.dim + 1), dtype=bool)

        for i in range(self.dim + 1):
            for j in range(self.dim + 1):
                for k in range(self.dim + 1):
                    s_i = self.grade_to_slice[i]
                    s_j = self.grade_to_slice[j]
                    s_k = self.grade_to_slice[k]

                    m = self.cayley[s_i, s_j, s_k]
                    gp_paths[i, j, k] = (m != 0).any()

        return gp_paths




from torch_ga.clifford import CliffordAlgebra
from torch_ga import GeometricAlgebra
import unittest as ut
import torch


from torch.testing import assert_close

class TestDualGeometricAlgebraMultiply(ut.TestCase):
    
    def assertTensorsApproxEqual(self, a, b, tolerance=1e-4):
        print(f"assertTensorsApproxEqual(a={a},b={b})")
        assert_close(a,b,atol=tolerance, rtol=0), "%s not equal to %s" % (a, b)
        
    def test_prod(self):
        pga_signature = [0, 1, 1, 1]
        pga_signature = [1, 1, -1]
        pga = CliffordAlgebra(pga_signature)
        
        # a = 3e01 + 5e02
        a = pga.embed(torch.tensor([3,5]),(3,4))
        b = pga.embed(torch.tensor([2,1]),(3,4))
                      
        self.assertTensorsApproxEqual(pga.product(a,b),pga.inner_product(a,b)+pga.outer_product(a,b))
    
        
        a = pga(a)
        b = pga(b)
        print(a,b)

        ga = GeometricAlgebra(pga_signature)
        
        pga = ga.to_cl()

        # a = 3e01 + 5e02
        a = pga.embed(torch.tensor([3,5]),(3,4))
        b = pga.embed(torch.tensor([2,1]),(3,4))
                      
        self.assertTensorsApproxEqual(pga.product(a,b),pga.inner_product(a,b)+pga.outer_product(a,b))


if __name__ == '__main__':
    ut.main()        